#include <inttypes.h>
#include <avr\signal.h>
#include <avr\eeprom.h>
#include <avr\pgmspace.h>
#include <avr\crc16.h>
#include <stdio.h>

#include "mmc_spi.h"

#define XTAL                16000000

#define LED_RED             PB0
#define LED_GREEN           PB1
#define LED_BLUE            PB2
#define LED_UNUSED          PB3

#define SER_SDA             PD0
#define SER_SCL             PD1
#define SER_EN              PE0

#define INT_COIL            PD2
#define INT_HALL            PD3

#define I2C_SDA             PD4
#define I2C_SCL             PD5
#define I2C_CS              PE2

#define SRAM_EN             PC7


#define LED_GREEN_ON()      PORTB &= ~(1 << LED_GREEN)
#define LED_GREEN_OFF()     PORTB |=  (1 << LED_GREEN)
#define LED_RED_ON()        PORTB &= ~(1 << LED_RED)
#define LED_RED_OFF()       PORTB |=  (1 << LED_RED)
#define LED_BLUE_ON()       PORTB &= ~(1 << LED_BLUE)
#define LED_BLUE_OFF()      PORTB |=  (1 << LED_BLUE)

#define _XRAM_                __attribute__ ((section (".noinit")))


int uartPut(char c) {

    if (c == '\n') uartPut('\r');
    while (!(UCSR0A & (1 << UDRE0)));
    UDR0 = c;
    return(0);
}





// teste externe 32Kb SRAM
uint8_t testSRAM(void) {

    uint8_t fail = 0;
    for (uint16_t addr = 0x0510; addr < 0x8500; addr++) {
      volatile uint8_t *ptr;
      (uint16_t)(ptr) = addr;
      *ptr = 0xAA;
      if (*ptr != 0xAA) fail = 1;
      *ptr = 0x55;
      if (*ptr != 0x55) fail = 1;
      *ptr = 0xFF;
      if (*ptr != 0xFF) fail = 1;
      *ptr = 0x00;
    }
    return(fail);
}

void printsector(uint8_t command, uint8_t result, int16_t size) {

    printf("MMC %.2x result: %x\n", command, result);

    if (result == 0x00) {
      int16_t pos = 0;
      uint8_t * data = MMC_Buffer;
      while (size > 0) {
        for (uint8_t i = 16; i > 0; i--) {
          printf("%.2x ", *data++);
        }
        data -= 16;
        printf(" : ");
        for (uint8_t i = 16; i > 0; i--) {
          char c = *data++;
          if (c < 0x20) c = '.';
          printf("%c", c);
        }
        printf("\n");
        size -= 16;
      }
    }
}

volatile uint8_t counter = 0;

SIGNAL(SIG_OVERFLOW1) {

    counter++;
}

void testMMC(void) {

    uint16_t sector = 0;

    TCCR1B = (1 << CS12);
    TIMSK = (1 << TOIE1);
    asm("sei");

    counter = 0;
    while (counter < 5) {
      mmcReadSector(sector++, MMC_Buffer);
    }

    printf("Test MMC: %d\n", sector);
//    uint8_t result = mmcReadSector(0, MMC_Buffer);
//    printsector(0x51, result, 512);

}


int main(void) {

    PORTB = (1 << LED_RED) | (0 << LED_GREEN) | (0 << LED_BLUE) | (1 << MMC_CS);
    DDRB = (1 << LED_RED) | (1 << LED_GREEN) | (1 << LED_BLUE) | (1 << MMC_CS) | (1 << MMC_SCLK) | (1 << MMC_MOSI);

    PORTD = (1 << SER_SDA) | (1 << SER_SCL) | (1 << INT_COIL) | (1 << INT_HALL) | (1 << I2C_SDA) | (1 << I2C_SCL);
    DDRD = 0;

    PORTE = (1 << SER_EN) | (1 << I2C_CS);
    DDRE = (1 << SER_EN) | (1 << I2C_CS);

    DDRC = (1 << SRAM_EN);
    MCUCR = (1 << SRE);
    SFIOR = (1 << XMM0);

    UCSR0B = (1 << TXEN0);
    UCSR0C = (1 << URSEL0) | (1 << UCSZ01) | (1 << UCSZ00);
    UBRR0H = 0;
    UBRR0L = 51;

    fdevopen(uartPut, 0, 0);

    printf("startup\n");
    printf("testing external SRAM:\n");
    if (testSRAM()) {
      LED_RED_ON();
      printf("fail\n");
    } else {
      printf("ok\n");
    }

    printf("testing MMC/SD card:\n");
    if (mmcInit()) {
      LED_RED_ON();
      printf("fail\n");
    } else {
      printf("ok\n");
      testMMC();
    }

    LED_BLUE_OFF();

    asm("sei");

    while (1) {
    }
}
