#include "mmc_spi.h"
#include <stdio.h>



uint8_t mmcReadBuffer(uint32_t address, uint8_t *buffer, uint16_t size) {


}


uint8_t mmcInit(void) {

    MMC_PORT |= (1 << MMC_CS);
    MMC_DDR |= (1 << MMC_CS) | (1 << MMC_SCLK) | (1 << MMC_MOSI);
    MMC_DDR &= ~(1 << MMC_MISO);
    SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0);
    SPSR = 0;

    uint8_t result = mmcCommandExpected(0x40, 0x0101);
    if (result <= 0x01) {
      for (uint8_t i = 20; i > 0; i--) {
        result = mmcCommand(0x41);
        if (result != 0x01) break;
      }
      if (result == 0x00) {
        SPCR = (1 << SPE) | (1 << MSTR);
        SPSR = (1 << SPI2X);
        #ifdef MMC_CALC_CRC7
        result = mmcCommandParam(0x7B, 1, 0x0000);  // activate CRC checking
        #endif
        if (result == 0) {
          result = mmcReadCSD(MMC_Buffer);

        }
      }
    }
    mmcDisable();
    return(result);
}


/*

  Follow the equivalent C code of assembler procedures in file mmc_spi_asm.S
  The assembler is more efficient as the compiled C source on WinAVR GCC.

#define mmcEnable()         MMC_PORT &= ~(1 << MMC_CS)
#define mmcDisable()        MMC_PORT |= (1 << MMC_CS)
#define mmcRead()           SPDR
#define mmcSend(data)       SPDR = data
#define mmcWait()           while (!(SPSR & (1 << SPIF)))


#ifdef MMC_CALC_CRC7
uint8_t mmcCRC7(uint8_t crc, uint8_t data) {

    crc ^= data;
    unsigned char poly = 0x89;
    for (unsigned char i = 8; i > 0; i--) {
      if ((crc & 0x80) != 0) crc ^= poly;
      crc <<= 1;
    }
    return(crc);
}
#endif

uint8_t mmcCommand(uint8_t command) {

    return(mmcCommandParam(command, 0, 0));
}

uint8_t mmcCommandExpected(uint8_t command, uint16_t expected) {

    return(mmcCommandParam(command, 0, expected));
}

uint8_t mmcCommandParam(uint8_t command, uint32_t address, uint16_t expected) {

    mmcDisable();
    mmcSend(data);

    uint8_t crc = 0;
#ifdef MMC_CALC_CRC7
    crc = mmcCRC7(crc, command);
    crc = mmcCRC7(crc, (uint8_t)(address >> 24));
    crc = mmcCRC7(crc, (uint8_t)(address >> 16));
    crc = mmcCRC7(crc, (uint8_t)(address >> 8));
    crc = mmcCRC7(crc, (uint8_t)(address));
    crc = crc | 1;
#else
    crc = 0x95;
#endif

    uint8_t result;
    uint8_t data = 0xFF;

    mmcWait();
    mmcEnable();
    for (uint8_t i = MMC_COMMAND_TRIALS; i > 0; i--) {
      mmcSend(command);
      mmcWait();
      mmcSend((uint8_t)(address >> 24));
      mmcWait();
      mmcSend((uint8_t)(address >> 16));
      mmcWait();
      mmcSend((uint8_t)(address >> 8));
      mmcWait();
      mmcSend((uint8_t)(address));
      mmcWait();
      mmcSend(crc);
      mmcWait();

      for (uint8_t j = MMC_RESPONSE_TRIALS; j > 0; j--) {
        mmcSend(data);
        mmcWait();
        result = mmcRead();
        if ((result & 0x80) == 0) break;
      }

      if ((result >= (uint8_t)(expected >> 8)) && (result <= (uint8_t)(expected))) break;

      mmcDisable();
      mmcSend(data);
      for (uint8_t j = MMC_COMMAND_DELAY -1; j > 0; j--) {
        mmcWait();
        mmcSend(data);
      }
      mmcWait();
      mmcEnable();
    }
    return(result);
}

uint8_t mmcReadData(uint16_t size, uint8_t *buffer) {
// the parameter order of size and buffer is choosen this way for better assembler optimization

    uint8_t result;
    uint8_t data = 0xFF;

    mmcSend(data);
    for (uint16_t i = MMC_READ_TRIALS -1; i > 0; i--) {
      mmcWait();
      result = mmcRead();
      mmcSend(data);
      if (result == 0xFE) goto DoRead;
    }
    return(0xFF);

  DoRead:;
    for (; size > 0; size--) {
      mmcWait();
      result = mmcRead();
      mmcSend(data);
     *buffer++ = result;
    }
    mmcWait();             // skip 16Bit CRC
    mmcRead();
    mmcSend(data);
    mmcWait();
    mmcRead();
    return(0);
}

uint8_t mmcReadSector(uint32_t sector, uint8_t *buffer) {

    uint8_t result = mmcCommandParam(0x51, sector * 512, 0x0000);
    if (result == 0x00) result = mmcReadData(buffer, 512);
    mmcDisable();
    return(result);
}

uint8_t mmcReadBuffer16(uint8_t command, uint8_t *buffer) {

    uint8_t result = mmcCommand(command);
    if (result == 0x00) result = mmcReadData(16, buffer);
    mmcDisable();
    return(result);
}

*/
