.nolist
#include "mmc_spi.inc"
.list

// some macros to easily take changes

.macro  mmcCS   enable             // Chip Select MMC
        .if (\enable)
        cbi     _SFR_IO_ADDR(MMC_PORT), MMC_CS
        .else
        sbi     _SFR_IO_ADDR(MMC_PORT), MMC_CS
        .endif
.endm

.macro  mmcWait
        sbis    _SFR_IO_ADDR(SPSR), SPIF
        rjmp    .-4
.endm

.macro  mmcRead reg                // read data
        in      \reg, _SFR_IO_ADDR(SPDR)
.endm

.macro  mmcSend reg                // send data
        out     _SFR_IO_ADDR(SPDR), \reg
.endm

        .section .text

#ifdef MMC_CALC_CRC7
mmcCRC:                            // update CRC7 in XL with polynom XH
        ldi     r25, 8
        sbrc    XL, 7
        eor     XL, XH
        lsl     XL
        dec     r25
        brne    mmcCRC +2
        ret
#endif

        .global  mmcCommand         // uint8_t mmcCommand(uint8_t command)
mmcCommand:
        clr     r23                 // expected = 0x0000
        clr     r22

        .global  mmcCommandExpected // uint8_t mmcCommand(uint8_t command, uint16_t expected)
mmcCommandExpected:                 // r25:r24               r25:r24           r23:r22
        movw    r18, r22            // address = 0x00000000
        clr     r23
        clr     r22
        clr     r21
mmcCommandParam_:
        clr     r20

        .global  mmcCommandParam    // uint8_t mmcCommandParam(uint8_t command, uint32_t address, uint16_t expected)
mmcCommandParam:                    // r25:r24                      r25:r24       r23:r22:r21:r20       r19:r18
                                    // don't use ZH:ZL !!
        mmcCS   0
        dec     r1                  // r1 == _zero_reg_ == always 0
        mmcSend r1

#ifdef MMC_CALC_CRC7
        ldi     XH, 0x89            // polynomial CRC7
        mov     XL, r24             // calculate CRC7 over command || address
        rcall   mmcCRC
        eor     XL, r23
        rcall   mmcCRC
        eor     XL, r22
        rcall   mmcCRC
        eor     XL, r21
        rcall   mmcCRC
        eor     XL, r20
        rcall   mmcCRC
        ori     XL, 1
#else
        ldi     XL, 0x95            // iff CRC7 checking inactive use CRC of command 0
#endif
        mov     r0, XL
        ldi     XL, MMC_COMMAND_TRIALS
        mmcWait
        mmcCS   1

mmcCommand1:
        mmcSend r24                // command
        mmcWait
        mmcSend r23                // address
        mmcWait
        mmcSend r22
        mmcWait
        mmcSend r21
        mmcWait
        mmcSend r20
        mmcWait
        mmcSend r0                 // CRC7
        ldi     r25, MMC_RESPONSE_TRIALS // read response, until MSB cleared
        mmcWait
mmcCommand2:
        mmcSend r1
        dec     r25
        mmcWait
        mmcRead XH
        sbrc    XH, 7
        brne    mmcCommand2
        cp      XH, r19           // check response with expected values
        brlo    mmcCommand3
        cp      r18, XH
        brsh    mmcCommand5

mmcCommand3:
        mmcCS   0
        ldi     r25, MMC_COMMAND_DELAY -1 // give card some cycles time
        mmcSend r1
mmcCommand4:
        mmcWait
        mmcSend r1
        dec     r25
        brne    mmcCommand4
        mmcWait
        mmcCS   1

        dec     XL
        brne    mmcCommand1

mmcCommand5:
        clr     r1
        mov     r24, XH
        clr     r25
        ret


        .global mmcReadData         // uint8_t mmcReadData(uint16_t size, uint8_t *buffer)
mmcReadData:                        // r25:r24               r25:r24          r23:r22
        movw    ZL, r22
mmcReadData_Z:
        ldi     r21, 0xFF
        mmcSend r21
        ldi     r23, hi8(MMC_READ_TRIALS)
        ldi     r22, lo8(MMC_READ_TRIALS)
mmcReadData1:
        mmcWait
        mmcRead r20
        mmcSend r21
        cpi     r20, 0xFE
        breq    mmcReadData2
        subi    r22, 1
        sbci    r23, 0
        brcc    mmcReadData1
        ldi     r24, 0xFF
        mmcWait
        rjmp    mmcReadData4

mmcReadData2:
        subi    r24, 1
        sbci    r25, 0
        brcs    mmcReadData3
        mmcWait
        mmcRead r20
        mmcSend r21
        st      Z+, r20
        rjmp    mmcReadData2

mmcReadData3:
        mmcWait                 // read 16bit CRC, ignore it
        mmcRead r20
        mmcSend r21
        mmcWait
        mmcRead r20
        clr     r24
mmcReadData4:
        clr     r25
        ret

        .global mmcReadSector   // uint8_t mmcReadSector(uint32_t sector, uint8_t *buffer)
mmcReadSector:                  // r25:r24               r25:r24:r23:r22    r21:r20
        movw    ZL, r20
        clr     r18
        clr     r19
        mov     r21, r22       // address = sector * 512
        mov     r22, r23
        mov     r23, r24
        add     r21, r21
        adc     r22, r22
        adc     r23, r23
        ldi     r24, 0x51
        rcall   mmcCommandParam_
        tst     r24
        brne    mmcDisable
        ldi     r25, hi8(512)
mmcReadSector1:
        rcall   mmcReadData_Z

        .global mmcDisable      // attention don't reorder
mmcDisable:
        mmcCS   0
        ldi     r23, 0xFF
        mmcSend r23
        mmcWait
        ret

        .global mmcReadBuffer16 // uint8_t mmcReadBuffer16(uint8_t command, uint8_t *buffer)
mmcReadBuffer16:                // r25:r24                     r25:r24        r23:r22
        movw    ZL, r22
        rcall   mmcCommand
        tst     r24
        brne    mmcDisable
        ldi     r24, 16
        rjmp    mmcReadSector1


